/**
   chat: a TCP/IP chatting program
   Copyright (C) 2006  Idan Nof

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Idan Nof can be reached via -

   email: idannof (at) hotmail (dot) com
   web  : http://www.geocities.com/idannof
   paper: 3 Brande st.,
		  Petach-Tikva, 49600,
		  Israel
*/
package Server;
import java.util.HashMap;
import java.util.Map;

/**
 * A database for holding users, session and the connection between them.
 * The database is synchronized where appropriate, and is safe for multi-threading.
 */
public final class DB {
	/**
	 * A matching between users (e.g: an IP Address) and their session. 
	 */
	private static Map sessionByUser = new HashMap();

	/**
	 * A matching between Names (e.g: strings) and their session.
	 */
	private static Map sessionByName = new HashMap();

	/**
	 * Retrieve a session by its name
	 * @param sessionName The name of the session to retrieve
	 * @return The session associated with the given name
	 */
	synchronized public static final Session getSessionByName(String sessionName) {
		return (Session) sessionByName.get(sessionName);
	}

	/**
	 * Add a given session to the database
	 * @param s The session to add
	 */
	synchronized static public final void addSession(Session s) {
		System.out.println("adding session: " + s);
		Session prevValue = (Session) sessionByName.put(s.getSessionName(), s);
		//assert prevValue == null : "session already exists";
		
		SessionWrapper newSessionThread = new SessionWrapper(s);
		s.setSessionWrapper(newSessionThread);
	}

	/**
	 * Get a session by a given user
	 * @param user A user
	 * @return The associated session
	 */
	synchronized static public final Session getSessionByUser(User user) {
		return (Session) sessionByUser.get(user);
	}

	/**
	 * Remove a given session from the database
	 * @param s The session to remove
	 */
	synchronized static public final void removeSession(Session s) {
		System.out.println("removing session: " + s);
		//assert s.getUsers().isEmpty() : "session not empty, cannot delete";
		Session prevValue = (Session) sessionByName.remove(s.getSessionName());
		//assert prevValue != null : "session does not exists";
	}

	/**
	 * Add a given user to a given session. 
	 * The method assumes the session is a valid registered session instance. 
	 * @param user The user to add
	 * @param session The session to add to
	 */
	synchronized static public final void addUser(User user, Session session) {
		System.out.println("adding: " + user + ", " + session);
		Map users = session.getUsers();
		//assert users != null : "this was supposed to be initialized";

		Session s = getSessionByName(session.getSessionName());
		//assert s != null : "unregistered session";

		// add user
		boolean isNewUser = users.containsKey(user);
		//assert !isNewUser : "user already exists";
		users.put(user, user);
		Object prevValue = sessionByUser.put(user, session);
		//assert prevValue == null : "inconsistent database";
	}

	/**
	 * Remove a user from a given session.
	 * @param user The use to remove
	 * @param session The session to remove the user from
	 */
	synchronized static public final void removeUser(User user, Session session) {
		System.out.println("removing: " + user + ", " + session);

		// remove from session members
		Map users = session.getUsers();
		User prevValue2 = (User) users.remove(user);
		//assert(
		//	prevValue2 != null) : "fatal: inconsistent request: user not found";

		// remove from user-session mappings
		Object prevValue = sessionByUser.remove(user);
		//assert prevValue
		//	!= null : "fatal: inconsistent request: session-user not found";
	}
}
