/**
   chat: a TCP/IP chatting program
   Copyright (C) 2006  Idan Nof

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Idan Nof can be reached via -

   email: idannof (at) hotmail (dot) com
   web  : http://www.geocities.com/idannof
   paper: 3 Brande st.,
		  Petach-Tikva, 49600,
		  Israel
*/
package Server;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.util.Map;

import Util.Config;
import Util.MyRandom;
import Util.UserIO;
import Util.Util;

/**
 * A thread responsible for handling all of the traffic of an open socket
 */
public final class SocketThread extends Thread {
	/** An output for the socket */
	private BufferedWriter sockOut;

	/**
	 * An input for the socket
	 */
	private BufferedReader sockIn;

	/**
	 * The user associated with this socket thread
	 */
	private User user;

	/**
	 * The socket associated with this socket thread.
	 * It is being initialized once in the constructor.
	 */
	private final Socket socket;

	/**
	 * A predicate that indicates whether we are after the open command
	 */
	private boolean isAfterOpenCommand = false;

	/**
	 * The session thread associated with this object
	 */
	private SessionWrapper sessionWrapper;

	/**
	 * A constructor that that associates a given socket with this object
	 * @param socket The socket to associate
	 */
	SocketThread(Socket socket) {
		this.socket = socket;
		doGetReaderWriter();
	}

	/**
	 * A procedure that sets the global 'sockIn' variable to wrap the socket InputStreamReader,
	 * and 'sockOut' to wrap the socket OutputStreamWriter.
	 */
	private final void doGetReaderWriter() {
		try {
			InputStream inputStream = socket.getInputStream();
			InputStreamReader inputStreamReader =
				new InputStreamReader(inputStream);
			sockIn = new BufferedReader(inputStreamReader);

			OutputStream outputStream = socket.getOutputStream();
			OutputStreamWriter outputStreamWriter =
				new OutputStreamWriter(outputStream);
			sockOut = new BufferedWriter(outputStreamWriter);

		} catch (IOException e) {
			System.out.println("Cannot get input and output streams.");
		}
	}

	/**
	 * Retrieve or create a user
	 * @param newUser The temporary user instance
	 * @param session The session to add the user to
	 */
	synchronized void doAddUser(User newUser, Session session) {
		boolean isNewUser = (null == DB.getSessionByUser(newUser));
		if (isNewUser) {
			DB.addUser(newUser, session);
			newUser.setSocketThread(this);
			setUser(newUser);
		}

		getSessionWrapper().synchronizeSession(session);
	}

	/**
	 * Retrieve or create a session
	 * @param newSession The temporary session instance
	 * @return The session retrieved or created
	 */
	synchronized Session doAddSession(Session newSession) {
		Session session = DB.getSessionByName(newSession.getSessionName());
		if (session == null) {
			DB.addSession(newSession);
			session = newSession;
		}
		setSessionThread(session.getSessionWrapper());
		return session;
	}

	/**
	 * Identify the user
	 * @return The user instance
	 */
	User doIdentifyUser() {
		String userName =
			new Integer(MyRandom.getRandom().nextInt(1000)).toString();
		User newUser = new User(socket.getInetAddress(), userName);

		return newUser;
	}

	/**
	 * A front-end to parse an open command
	 * @param line The open command string
	 * @return The parsed session
	 */
	Session parseOpenCommand(String line) {
		Session newSession = UserIO.parseOpenCommand(line);
		if (newSession == null) {
			Util.sendLine("open command expected", sockOut);
			return null;
		}

		return newSession;
	}

	/**
	 * Handle an open command
	 * @param line The open command string
	 */
	private void handleOpenCommand(String line) {
		Session newSession = parseOpenCommand(line);
		if (newSession == null)
			return;
		User newUser = doIdentifyUser();
		Session session = doAddSession(newSession);
		doAddUser(newUser, session);

		isAfterOpenCommand = true;
		getSessionWrapper().newMessage(
			this,
			"user " + getUser() + " has joined the group");
	}

	/**
	 * Handle a given input line from the network;
	 * If we have identified a user-session matching, send it to the session member,
	 * otherwise, take care of authentication and book-keeping.
	 * @param line The line to handle
	 */
	private final void handleLine(String line) {
		if (isAfterOpenCommand) {
			getSessionWrapper().newMessage(this, line);
		} else {
			handleOpenCommand(line);
		}
	}

	/**
	 * A front-end for this thread;
	 * 1. Infinitely block until a line is recieved over the network.
	 * 2. Handle the line.
	 */
	public void run() {
		int retries = 0;

		while (true) {
			String line = Util.recvLine(sockIn);
			if (line == null) {
				System.out.println(socket + ": unexpected end of connection");
				retries++;
				if (retries == Config.max_socket_retries)
					break;
				else
					continue;
			}
			System.out.println(
				socket.getInetAddress()
					+ ", user="
					+ user
					+ ": '"
					+ line
					+ "'");
			handleLine(line);
		}

		finish();
	}

	private void finish() {
		System.out.println(socket + ": aborting connection ");
		try {
			sockIn.close();
			sockOut.close();
			socket.close();
		} catch (IOException e) {
			System.out.println("cannot gracefully close socket.");
		}

		if (user != null) {
			// doRemoveUser
			Session session = DB.getSessionByUser(user);
			DB.removeUser(user, session);

			// doRemoveSession
			Map users = session.getUsers();
			if (users.isEmpty()) {
				DB.removeSession(session);
			}
		}
		System.out.println(this +": quitting");
	}

	/**
	 * Returns the user associated with this socket thread
	 * @return the user associated with this socket thread
	 */
	public User getUser() {
		return user;
	}

	/**
	 * Sets the user associated with this socket thread
	 * @param user The user to associate with this socket thread
	 */
	private void setUser(User user) {
		this.user = user;
	}

	/**
	 * Returns a BufferedWriter output for the associated socket
	 * @return a BufferedWriter output for the associated socket
	 */
	public BufferedWriter getSockOut() {
		return sockOut;
	}

	/**
	 * Returns the session thread associated with this object
	 * @return the session thread associated with this object
	 */
	private SessionWrapper getSessionWrapper() {
		return sessionWrapper;
	}

	/**
	 * Sets the session thread associated with this object
	 * @param thread The thread to save
	 */
	private void setSessionThread(SessionWrapper thread) {
		sessionWrapper = thread;
	}

}
